/* VirtuinoSE example: Send text to VirtuinoSE app & Receive Text from VirtuinoSE app
 * Supported boards: Arduino Uno or Nano + ESP8266 module 
 * Created by Ilias Lamprou
 * Updated Oct 27 2018
 * More examples: http://virtuino.com/index.php/virtuino-se/20-virtuino-se-code-examples
 */
// ------ VirtuinoSE V Memory Table
//  V1 memory  -> Text 1
//  V2 memory  -> Text 2

//--------------- USER SETTINGS ---------------------------------------
const char* ssid = "WIFI NETWORK";           // WIFI network SSID or AP SSID
const char* password = "PASSWORD";           // WIFI network PASSWORD

#define VIRTUINO_KEY  "1234"                           // Virtuino KEY, only the requests that include this key will be acceptable
                                              // disable the line if you don't want to use a key                          
#include <SoftwareSerial.h>
SoftwareSerial espSerial =  SoftwareSerial(2,3);      // arduino RX pin=2  arduino TX pin=3    connect the arduino RX pin to esp8266 module TX pin   -  connect the arduino TX pin to esp8266 module RX pin

int port =8000;

String textHolder_1="Hello";
String textHolder_2="Virtuino";

unsigned long timeStrored=0;
boolean debug=true;

//============================================================== connectToWiFiNetwork
void connectToWiFiNetwork(){
    Serial.println("Connecting to "+String(ssid));
    clearESP_buffer(1000);
    espSerial.println("AT+GMR");
    waitForResponse("OK",1000);
    
    espSerial.println("AT+CWMODE=1");  // configure as client
    waitForResponse("OK",1000);
    
    //--- connect
    espSerial.print("AT+CWJAP=\"");
    espSerial.print(ssid);
    espSerial.print("\",\"");
    espSerial.print(password);
    espSerial.println("\"");
    waitForResponse("OK",10000);
     
    espSerial.println("AT+CIFSR");           // get ip address
    waitForResponse("OK",1000);
    
    espSerial.println("AT+CIPMUX=1");        // configure for multiple connections   
    waitForResponse("OK",1000);
    
    espSerial.print("AT+CIPSERVER=1,");
    espSerial.println(port);
    waitForResponse("OK",1000);
   
}

//=================================================================
void initAccessPoint(){
      if (debug) Serial.println( "Init as AP...");
      clearESP_buffer(1000);
      espSerial.println("AT+GMR");
      waitForResponse("OK",1000);

      espSerial.println("AT+CWMODE=2");  // configure as client
      waitForResponse("OK",1000);
    
     espSerial.println("AT+CIFSR");
     waitForResponse("OK",1000);

     espSerial.print("AT+CWSAP=\"");
     espSerial.print(ssid);
     espSerial.print("\",\"");
     espSerial.print(password);
     espSerial.println("\",2,3");
     waitForResponse("OK",5000);
     
     espSerial.println("AT+CIPMUX=1");         // configure for multiple connections   
     waitForResponse("OK",2000);
    
     espSerial.print("AT+CIPSERVER=1,");
     espSerial.println(port);
     waitForResponse("OK",1000);           
}


//============================================================== setup
//==============================================================
void setup() {
  Serial.begin(9600);
  espSerial.begin(9600);
  connectToWiFiNetwork();      //enable this line to connect the module your wifi network
  //initAccessPoint();         //enable this line to use the esp as AP  
  
  pinMode(13,OUTPUT);        
  }

//============================================================== onCommandReceived
//==============================================================
/* This function is called every time VirtuinoSE app sends a request to server 
 * The 'variableIndex' is the V memory of Virtuino app
 * The 'valueAsText' is the value that has sent from the app
 * If the 'valueAsText" is ? you have to return the value that you want to be displayed on Virtuino panel
 */
 
String onCommandReceived(int variableIndex, String valueAsText){
  if (debug) Serial.println("New command: variableIndex="+String(variableIndex)+"   valueAsText="+valueAsText);
 //---------------------  Read Write texts
        if (variableIndex==1) {                      // V1 memory variable contains a text or a value
          if (valueAsText=="?") return urlencode(&textHolder_1);   // encode and decode the text to avoid communication issues
          else  textHolder_1=urldecode(&valueAsText);
        }
        else if (variableIndex==2) {                 // V2 memory variable contains a text or a value
          if (valueAsText=="?") return urlencode(&textHolder_2);
           else textHolder_2=urldecode(&valueAsText);
        }
 
   
  return "";
}


//============================================================== loop
//==============================================================
void loop() {
  virtuinoRun();

  
}












//================== VirtuinoSE Library (includend in the sketch)=============
//=============================================================================
//=============================================================================

/* Only the next two functions are nedded to communicate with Virtuino 
 * Avoid to make changes to the code below
 */


#define WRONG_KEY_MESSAGE  "Wrong Key"  
#define WELLCOME_MESSAGE  "Hello Virtuino"  

//============================================================== virtuinoRun
//==============================================================
void virtuinoRun(){
     if(espSerial.available()){
       delay(1);
       String espBuffer = espSerial.readStringUntil('\r');
          if (debug) Serial.print(espBuffer);
          int pos = espBuffer.indexOf("+IPD,");
          if (pos>=0){                                        // check for GET command from Virtuino app
              clearESP_buffer(100);                           // can be higher than 100
              int connectionId = espBuffer.charAt(pos+5)-48;  // get connection ID
              pos = espBuffer.indexOf("GET /");
              if (pos!=-1){                                   // We have a GET message
                   espBuffer.remove(0,pos+5);                 // Clear the characters GET / 
                   int k= espBuffer.indexOf(" HTTP/");       
                   if (k>0) espBuffer.remove(k);                       // Clear the tail of the line
                   String responseData = getResponseData(&espBuffer);  // The espBuffer contains the Virtuino data
                                                                       
                    // Create HTTP Header and data
                    String httpResponse = "HTTP/1.1 200 OK\r\nContent-Type: text/html; charset=UTF-8\r\nContent-Length: ";
                    httpResponse += responseData.length();
                    httpResponse += "\r\nConnection: close\r\n\r\n";
                    httpResponse += responseData;
                    httpResponse += (" ");
                    //---Send response to Virtuino SE
                    String cipSend = "AT+CIPSEND=";
                    cipSend += connectionId;
                    cipSend += ",";
                    cipSend += httpResponse.length();
                    cipSend += "\r\n";
                    for (int i=0;i<cipSend.length();i++) espSerial.write(cipSend.charAt(i));
                    if (waitForResponse(">",2000)) {
                        for (int i=0;i<httpResponse.length();i++) {
                            espSerial.write(httpResponse.charAt(i));
                            if (debug) Serial.print(httpResponse.charAt(i));
                        }
                    }
              }
          }
     }

}

//============================================================== prepareResponse
//==============================================================
String getResponseData(String* commands){
   
   //---check the KEY
   #ifdef VIRTUINO_KEY 
      String virtuinoKey="";
      int keyPosStart = commands->indexOf("&KEY=");
      if (keyPosStart!=-1) {
        int keyPosEnd = commands->indexOf('&',keyPosStart+5);
        if (keyPosEnd!=-1) virtuinoKey= commands->substring(keyPosStart+5,keyPosEnd);
        else virtuinoKey= commands->substring(keyPosStart);
      }
      if (!virtuinoKey.equals(VIRTUINO_KEY)) return WRONG_KEY_MESSAGE;
   #endif
   
  //--- read all virtuino commands from request data
  String commandResponse="";
  boolean checkNext=true;
  int cStart=-1;
  while (checkNext){
    cStart=commands->indexOf('&',cStart+1);
    if (cStart!=-1) {
      int cEnd=commands->indexOf('&',cStart+1);
      String oneCommand="";
      if (cEnd<0) oneCommand=commands->substring(cStart+1); 
      else oneCommand=commands->substring(cStart+1,cEnd); 
      if (debug) Serial.println("cmd: "+oneCommand);
      char commandType = oneCommand.charAt(0);
      if (commandType=='C') return "&CNT="+String(WELLCOME_MESSAGE);               // Virtuino sends the command &C=? to check the connection status   
      else if ((commandType=='V') || (commandType=='T')) {
        int equalCharPosition = oneCommand.indexOf('=');
        int variableIndex  = oneCommand.substring(1,equalCharPosition).toInt(); 
        String valueString = oneCommand.substring(equalCharPosition+1); 
        String returnedValue = onCommandReceived(variableIndex,valueString);  
       if (returnedValue.length()>0) {
          commandResponse+='&';
          commandResponse+=commandType;
          commandResponse+=String(variableIndex);
          commandResponse+='=';
          commandResponse+=returnedValue;
        }
      }
    }
    else checkNext=false;   // the last command has checked
  }
  if (debug) Serial.println("Response="+commandResponse );
   if (commandResponse.length()==0) commandResponse="OK";       
  return commandResponse;
}

//======================================================================================== clearESP_buffer
//========================================================================================
void clearESP_buffer(int timeout){
    long t=millis();
    char c;
    while (t+timeout>millis()){
      if (espSerial.available()){
        c=espSerial.read();
        if (debug) Serial.print(c);
      }
    }
   }

//======================================================================================== waitForResponse
//========================================================================================
boolean waitForResponse(String target1,  int timeout){
    String data;
    char a;
    unsigned long start = millis();
    boolean rValue=false;
    while (millis() - start < timeout) {
        while(espSerial.available() > 0) {
            a = espSerial.read();
            if (debug) Serial.print(a);
            if(a == '\0') continue;
            data += a;
        }
        if (data.indexOf(target1) != -1) {
            rValue=true;
            break;
        } 
    }
    return rValue;
}

 //======= ENCODE - DECODE Library (included in the sketch) ========
 //=================================================================
 
 //================================================================ urldecode
String urldecode(String* str){
   String encodedString="";
    char c;
    char code0;
    char code1;
    for (int i =0; i < str->length(); i++){
        c=str->charAt(i);
      if (c == '+'){
        encodedString+=' ';  
      }else if (c == '%') {
        i++;
        code0=str->charAt(i);
        i++;
        code1=str->charAt(i);
        c = (h2int(code0) << 4) | h2int(code1);
        encodedString+=c;
      } else{
         encodedString+=c;  
      }
      yield();
    }
    
   return encodedString;
}
//================================================================ urlencode
String urlencode(String* str){
   String encodedString="";
    char c;
    char code0;
    char code1;
    char code2;
    for (int i =0; i < str->length(); i++){
      c=str->charAt(i);
      if (c == ' '){
        encodedString+= '+';
      } else if (isalnum(c)){
        encodedString+=c;
      } else{
        code1=(c & 0xf)+'0';
        if ((c & 0xf) >9){
            code1=(c & 0xf) - 10 + 'A';
        }
        c=(c>>4)&0xf;
        code0=c+'0';
        if (c > 9){
            code0=c - 10 + 'A';
        }
        code2='\0';
        encodedString+='%';
        encodedString+=code0;
        encodedString+=code1;
        //encodedString+=code2;
      }
      yield();
    }
    return encodedString;
}
//================================================================ h2int
unsigned char h2int(char c){
    if (c >= '0' && c <='9'){
        return((unsigned char)c - '0');
    }
    if (c >= 'a' && c <='f'){
        return((unsigned char)c - 'a' + 10);
    }
    if (c >= 'A' && c <='F'){
        return((unsigned char)c - 'A' + 10);
    }
    return(0);
}

  
//----------------------------------------------------------------------------
/* Virtuino GET request to upload values of V1 and V4 memories to board  (KEY=1234)
 *  GET /&KEY=1234&V1=12.3&V4=1
 *  Response: OK   
 *  
 * Virtuino GET request to read values from board (KEY=1234)
 *  GET /&KEY=1234&V2=?&V3=?
 *  Response: &V2=123&V3=1
 *  
 *  * Virtuino GET request to upload and read values from board (KEY=1234)
 *  GET /&KEY=1234&V2=45.6&V3=?
 *  Response: &V3=1
 *  
 *  Virtuino GET request to check the connection  (KEY=1234)
 *  GET /&KEY=1234&CNT=?
 *  Response: &CNT=Hello Virtuino I am the ESP8266 board
 *  
 * Virtuino GET request to upload value as TEXT from memory V1 to board  (KEY=1234)
 *  GET /&KEY=1234&T1=Hello
 *  Response: OK 

 * Virtuino GET request to read value a TEXT from board  (KEY=1234)
 *  GET /&KEY=1234&T1=?
 *  Response &T1=the text you want to send 
 *  or
 *  GET /&KEY=1234&V1=?
 *  Response &V1=the text you want to send 
 *  
 *  All the texts that are sent from Virtuino are encoded 
 *  All the texts that received from Virtuino will be decoded
 */



